/* Copyright 2004, 2005, 2006 Acegi Technology Pty Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.security.openid;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.openid4java.association.AssociationException;
import org.openid4java.consumer.ConsumerException;
import org.openid4java.discovery.DiscoveryException;
import org.openid4java.discovery.DiscoveryInformation;
import org.openid4java.message.AuthRequest;
import org.openid4java.message.MessageException;
import org.openid4java.message.ParameterList;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.security.openid.google.ConsumerHelperFactory;
import org.springframework.security.openid.google.DefaultConsumerHelperFactory;
import org.springframework.security.openid.google.GoogleAxFetchListFactory;

import com.google.step2.AuthRequestHelper;
import com.google.step2.AuthResponseHelper;
import com.google.step2.ConsumerHelper;
import com.google.step2.Step2;
import com.google.step2.VerificationException;
import com.google.step2.discovery.IdpIdentifier;
import com.google.step2.openid.ui.UiMessageRequest;

/**
 * @author Ray Krueger
 * @author Luke Taylor
 */
public class OpenID4JavaConsumer implements OpenIDConsumer, InitializingBean, ApplicationContextAware {
	
    private static final String DISCOVERY_INFO_KEY = DiscoveryInformation.class.getName();

    //~ Instance fields ================================================================================================

    protected final Log logger = LogFactory.getLog(getClass());

    private final AxFetchListFactory attributesToFetchFactory;
    
    private ApplicationContext applicationContext;
    
    private ConsumerHelper consumerHelper; 
    
    private List<OpenIDAttribute> attributesToFetch = Collections.emptyList();
    
    private List<OpenIdAuthenticationHandler> authenticationHandlers = null;

    //~ Constructors ===================================================================================================

    public OpenID4JavaConsumer() throws ConsumerException {
        this(new GoogleAxFetchListFactory());
    }
    
    public OpenID4JavaConsumer(AxFetchListFactory attributesToFetchFactory) throws ConsumerException {
    	this.attributesToFetchFactory = attributesToFetchFactory;
    }

    //~ Methods ========================================================================================================
    
    public String beginConsumption(HttpServletRequest req, String identityUrl, String returnToUrl, String realm) 
    		throws OpenIDConsumerException {
    	
    	IdpIdentifier openId = new IdpIdentifier(identityUrl);
    	AuthRequestHelper helper = consumerHelper.getAuthRequestHelper(openId, returnToUrl);
    	
		if (attributesToFetchFactory != null 
				&& !(attributesToFetchFactory instanceof NullAxFetchListFactory)) {
			attributesToFetch = attributesToFetchFactory.createAttributeList(identityUrl);
		}
    	for (OpenIDAttribute attribute : attributesToFetch) {
    		helper.requestAxAttribute(attribute.getName(), attribute.getType(), attribute.isRequired());
    	}
    	
        HttpSession session = req.getSession();
        try { 
	        AuthRequest authReq = helper.generateRequest();
	        authReq.setRealm(realm);
	        
	        UiMessageRequest uiExtension = new UiMessageRequest();
	        uiExtension.setIconRequest(true);
	        authReq.addExtension(uiExtension);

	        session.setAttribute(DISCOVERY_INFO_KEY, helper.getDiscoveryInformation());
	        return authReq.getDestinationUrl(true);
        } catch (MessageException e) {
        	throw new OpenIDConsumerException("Error processing ConsumerManager authentication", e);
        } catch (DiscoveryException e) {
        	throw new OpenIDConsumerException("Error processing ConsumerManager authentication", e);
        } catch (ConsumerException e) {
        	throw new OpenIDConsumerException("Error processing ConsumerManager authentication", e);
        }
    }

    public OpenIDAuthenticationToken endConsumption(HttpServletRequest request) throws OpenIDConsumerException {
		// extract the parameters from the authentication response
		// (which comes in as a HTTP request from the OpenID provider)
    	ParameterList openIdResponse = Step2.getParameterList(request);
    	
    	// retrieve the previously stored discovery information
    	DiscoveryInformation discovered = (DiscoveryInformation) request.getSession().getAttribute(DISCOVERY_INFO_KEY);
        if (discovered == null) {
            throw new OpenIDConsumerException("DiscoveryInformation is not available. Possible causes are lost session or replay attack");
        }
        request.getSession().removeAttribute(DISCOVERY_INFO_KEY);

        // extract the receiving URL from the HTTP request
        String receivingURL = Step2.getUrlWithQueryString(request);
        
        String queryString = request.getQueryString();
        if (queryString != null && queryString.length() > 0) {
        	receivingURL += "?" + request.getQueryString();
        }
        
        AuthResponseHelper authResponse = null; 
        try {
			authResponse = consumerHelper.verify(receivingURL, openIdResponse, discovered);
	        if (authResponse.getAuthResultType() == AuthResponseHelper.ResultType.AUTH_SUCCESS) {
	        	List<OpenIDAttribute> attributes = fetchAxAttributes(authResponse, request);
	        	triggerAuthenticationHandlers(
	        		request, authResponse.getClaimedId().getIdentifier(), attributes);
	        	
	        	return new OpenIDAuthenticationToken(OpenIDAuthenticationStatus.SUCCESS, 
	        		authResponse.getClaimedId().getIdentifier(), authResponse.getStatusMsg(), attributes);
	        }			
		} catch (MessageException e) {
			throw new OpenIDConsumerException("Error verifying openid response", e);
		} catch (AssociationException e) {
			throw new OpenIDConsumerException("Error verifying openid response", e);
		} catch (DiscoveryException e) {
			throw new OpenIDConsumerException("Error verifying openid response", e);
		} catch (VerificationException e) {
			throw new OpenIDConsumerException("Error verifying openid response", e);
		}
        
        return new OpenIDAuthenticationToken(OpenIDAuthenticationStatus.FAILURE, 
        	authResponse.getClaimedId().getIdentifier(), authResponse.getStatusMsg(), new ArrayList<OpenIDAttribute>());
    }

	private List<OpenIDAttribute> fetchAxAttributes(AuthResponseHelper authResponse, HttpServletRequest request) {
		List<OpenIDAttribute> result = new ArrayList<OpenIDAttribute>(attributesToFetch.size()); 
		for (OpenIDAttribute attribute : attributesToFetch) {
			String value = authResponse.getAxFetchAttributeValue(attribute.getName());
			result.add(new OpenIDAttribute(
				attribute.getName(), attribute.getType(), Arrays.asList(value)));
		}
		return result; 
	}
	
	private void triggerAuthenticationHandlers(HttpServletRequest request, 
			String identifier, List<OpenIDAttribute> attributes) {
		
		if (authenticationHandlers != null) {
			for (OpenIdAuthenticationHandler handler : authenticationHandlers) {
				handler.onOpenIdAuthenticationSuccess(request, identifier, attributes);
			}
		}
	}
	
	public ConsumerHelper getConsumerHelper() {
		return consumerHelper;
	}

	public void afterPropertiesSet() throws Exception {
		if (consumerHelper == null) {
			ConsumerHelperFactory factory = null;
			try {
				factory = applicationContext.getBean(ConsumerHelperFactory.class);
			} catch (NoSuchBeanDefinitionException ex) {
				factory = new DefaultConsumerHelperFactory();
			}
			consumerHelper = factory.createConsumerHelper();
		}
		this.authenticationHandlers = new ArrayList<OpenIdAuthenticationHandler>(
			applicationContext.getBeansOfType(OpenIdAuthenticationHandler.class).values());
	}

	public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
		this.applicationContext = applicationContext;
	}
}