/**
 * Copyright 2013 Rabbitsoft
 * Clinked
 */
package org.springframework.security.openid.memcache;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.spy.memcached.AddrUtil;
import net.spy.memcached.MemcachedClient;

import org.openid4java.association.Association;
import org.openid4java.consumer.ConsumerAssociationStore;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** 
 * Stores all associations in memcached server. 
 * 
 * @author see AUTHORS.txt
 */
public class MemcacheConsumerAssociationStore implements ConsumerAssociationStore {
	/** 
	 * {@link Logger}.
	 */
	private final Logger logger = LoggerFactory.getLogger(getClass());
	
	/** 
	 * Key used to store the whole association store.
	 */
	private final String storeKey = "MemcacheConsumerAssociationStore";
	
	/** 
	 * List of memcached server addresses.
	 */
	private List<String> addressList;
	
	/** 
	 * @param addressList {@link #addressList}
	 */
	public MemcacheConsumerAssociationStore(List<String> addressList) {
		this.addressList = addressList;
	}
	
	/** 
	 * @param addressList {@link #addressList}
	 */
	public MemcacheConsumerAssociationStore(String addressList) {
		this(Arrays.asList(addressList.split("(?:\\s|,)+")));
	}
	
	/** 
	 * Creates a new client. 
	 * 
	 * @return memcached client
	 * @throws IOException when fails
	 */
	private MemcachedClient getClient() throws IOException {
		return new MemcachedClient(AddrUtil.getAddresses(addressList));
	}
	
	/** 
	 * Used to perform an action in memcached with pre-loaded association store. 
	 * The store is automatically updated when modified in the handler.
	 * 
	 * @author see AUTHORS.txt
	 */
	interface AssociationHandler {
		
		/** 
		 * Do something on the store.
		 * 
		 * @param store store
		 * @return object or null
		 */
		Object handle(final Map<String, Map<String, Association>> store);
	}
	
	/** 
	 * Pre-load the store from memcached server and call the handler to 
	 * perform action on store.
	 * 
	 * @param handler handler
	 * @return optional object to return
	 */
	@SuppressWarnings("unchecked")
	private Object handleAssociationOperation(AssociationHandler handler) {
		MemcachedClient client = null;
		try { 
			client = getClient();
			Map<String, Map<String, Association>> store = 
				(Map<String, Map<String, Association>>) client.get(storeKey);
			
			if (store == null) {
				store = new HashMap<String, Map<String, Association>>();
			} else {
				removeExpired(store);
			}
			
			Object result = handler.handle(store);
			client.set(storeKey, 0, store);
			
			return result;
		} catch (IOException ex) {
			logger.error("Failed to do association operation:", ex);
		} finally {
			if (client != null) {
				client.shutdown();
			}
		}
		return null;
	}
	
	@Override
	public void save(final String opUrl, final Association association) {
		handleAssociationOperation(new AssociationHandler() {
			@Override
			public Object handle(Map<String, Map<String, Association>> store) {
				Map<String, Association> handleMap = store.get(opUrl);
		        if (handleMap == null) {
		            handleMap = new HashMap<String, Association>();
		        }
		        String handle = association.getHandle();
		        
		        logger.debug("Adding association to the memcache store: " + handle + " with OP: " + opUrl);
		        handleMap.put(handle, association);
		        
		        store.put(opUrl, handleMap);
				return null;
			}
		});
	}

	@Override
	public Association load(final String opUrl, final String handle) {
		return (Association) handleAssociationOperation(new AssociationHandler() {
			@Override
			public Object handle(Map<String, Map<String, Association>> store) {
				Map<String, Association> handleMap = (Map<String, Association>) store.get(opUrl);
		        if (handleMap != null) {
		            if (handleMap.containsKey(handle)) {
		                return handleMap.get(handle);
		            }
		        }
				return null;
			}
		});
	}

	@Override
	public Association load(final String opUrl) {
		return (Association) handleAssociationOperation(new AssociationHandler() {
			@Override
			public Object handle(Map<String, Map<String, Association>> store) {
				Map<String, Association> handleMap = (Map<String, Association>) store.get(opUrl);
		        if (handleMap != null) {
		        	Iterator<String> handles = handleMap.keySet().iterator();
		        	Association latest = null;
		        	while (handles.hasNext()) {
		        		String handle = (String) handles.next();
		        		Association association = (Association) handleMap.get(handle);
		        		if (latest == null 
		        				|| latest.getExpiry().before(association.getExpiry())) {
		        			latest = association;
		        		}
		        	}
		        	return latest;
		        }
				return null;
			}
		});
	}

	@Override
	public void remove(final String opUrl, final String handle) {
		handleAssociationOperation(new AssociationHandler() {
			@Override
			public Object handle(Map<String, Map<String, Association>> store) {
				Map<String, Association> handleMap = (Map<String, Association>) store.get(opUrl);
		        if (handleMap != null) {
		        	logger.info("Removing association: {} with OP: {}", handle, opUrl);
		        	
		        	handleMap.remove(handle);
		        	if (handleMap.size() == 0) {
		        		store.remove(opUrl);
		        	}
		        }
		        return null;
			}
		});	
	}
	
	/** 
	 * Removes any expired associations in store. 
	 * 
	 * @param store store
	 */
    private synchronized void removeExpired(Map<String, Map<String, Association>> store) {
    	
    	Set<String> opToRemove = new HashSet<String>();
    	for (String opUrl : store.keySet()) {
    		Map<String, Association> handleMap = store.get(opUrl);
    		
    		Set<String> handleToRemove = new HashSet<String>();
    		for (String handle : handleMap.keySet()) {
    			  Association association = handleMap.get(handle);
    			  if (association.hasExpired()) {
    				  handleToRemove.add(handle);
    			  }
    		}
    		for (String handle : handleToRemove) {
                logger.info("Removing expired association: " + handle + " with OP: " + opUrl);
                handleMap.remove(handle);    			
    		}
    		
            if (handleMap.size() == 0) {
                opToRemove.add(opUrl);
            }
    	}
    	for (String opUrl : opToRemove) {
    		store.remove(opUrl);
    	}
    }
}
