package org.springframework.security.openid.redis;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import redis.clients.jedis.Jedis;
import redis.clients.jedis.JedisPool;
import redis.clients.jedis.JedisPoolConfig;

public class RedisHelper {
	/** 
	 * {@link Logger}.
	 */
	private final Logger logger = LoggerFactory.getLogger(getClass());
	
	private JedisPool jedisPool;
	
	public RedisHelper(JedisPoolConfig config, String host) {
		jedisPool = new JedisPool(config, host);
	}
	
	public void destroy() { 
		jedisPool.destroy();
	}
	
	public Object get(final String keyName) { 
		return execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				return jedis.get(keyName.getBytes());
			}
		});		
	}
	
	public String set(final String key, final Object value) {
		return (String) execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				return jedis.set(key.getBytes(), serialize(value));
			}
		});			
	}
	
	public void expire(final String key, final int seconds) {
		execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				jedis.expire(key, seconds);
				return null;
			}
		});			
	}
	
	public void hset(final String hashName, final String key, final Object value) {
		execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				jedis.hset(hashName.getBytes(), key.getBytes(), serialize(value));
				return null;
			}
		});
	}
	
	public void hdel(final String hashName, final String key) {
		execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				jedis.hdel(hashName.getBytes(), key.getBytes());
				return null;
			}
		});
	}
	
	@SuppressWarnings("unchecked")
	public List<Object> hvals(final String hashName) {
		return (List<Object>) execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				List<Object> result = new ArrayList<Object>();
				for (byte[] data : jedis.hvals(hashName.getBytes())) {
					result.add(deserialize(data));
				}
				return result;
			}
		});
	}	
	
	public Object hget(final String hashName, final String key) {
		return execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				byte[] data = jedis.hget(hashName.getBytes(), key.getBytes());
				if (data == null) {
					return null;
				}
				return deserialize(data);
			}
		});
	}
	
	@SuppressWarnings("unchecked")
	public Collection<String> keys(final String prefix) {
		return (Collection<String>) execute(new JedisCallback() {
			@Override
			public Object execute(Jedis jedis) {
				return jedis.keys(prefix);
			}
		});
	}
	
	public interface JedisCallback { 
		Object execute(Jedis jedis);
	}	
	
	protected Object execute(JedisCallback callback) {
		Jedis jedis = jedisPool.getResource();
		try { 
			return callback.execute(jedis);
		} finally { 
			jedisPool.returnResource(jedis);
		}
	}
	
	protected byte[] serialize(Object object) {
		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        ObjectOutputStream out = null;
		try {
			out = new ObjectOutputStream(outputStream);
	        out.writeObject(object);
	        
	        out.close();
		} catch (Exception ex) {   
			logger.error("Failed to serialize object: {}", ex.getMessage());
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException ex) {
				}
			}
			try {
				outputStream.close();
			} catch (IOException e) {
			}
		}
		return outputStream.toByteArray();
	}
	
	protected  Object deserialize(byte[] data) {
		ByteArrayInputStream inputStream = new ByteArrayInputStream(data);
		ObjectInputStream objectInputStream = null;
		try {
			objectInputStream = new ObjectInputStream(inputStream);
			return objectInputStream.readObject();
		} catch (Exception ex) {
			logger.error("Failed to deserialize object: {}", ex.getMessage());
		} finally { 
			if (objectInputStream != null) {
				try {
					objectInputStream.close();
				} catch (IOException ex) {
				}
			}
			try {
				inputStream.close();
			} catch (IOException e) {
			}
		}
		return null;
	}
}
